//-----------------------------------------------------------------------------
// Copyright (c) 2008-2011 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;
using Dhpoware;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace XNAShadowMapping
{
    /// <summary>
    /// This Windows XNA application demonstrates shadow mapping for a single
    /// directional light source. Simple shadow mapping produces very aliased
    /// shadows. This demo uses percentage-closer filtering with a 2x2 and a
    /// 3x3 kernel to soften the shadows to reduce their aliased appearance.
    /// The shadow mapping effect file that accompanies this demo targets
    /// shader model 2.0 cards. Consequently shadows still appear aliased even
    /// with percentage-closer filtering enabled. More advanced algorithms
    /// exist that produces far superior soft shadows. However such algorithms
    /// typically require shader model 3.0 or higher.
    /// </summary>
    public class Demo : Microsoft.Xna.Framework.Game
    {
        private static void Main()
        {
            using (Demo demo = new Demo())
            {
                demo.Run();
            }
        }

        private struct Camera
        {
            public const float DOLLYING_SPEED = 0.10f;
            public const float ROTATION_SPEED = 0.50f;
            public const float MOUSEWHEEL_SPEED = 2.0f;
            public const float TRACKING_SPEED = 0.02f;

            public float fovy;
            public float znear;
            public float zfar;
            public float offset;
            public float maxZoom;
            public float minZoom;
            public Vector2 rotation;
            public Vector2 translate;
            public Vector3 position;
            public Vector3 target;
            public Vector3 viewDir;
            public Quaternion orientation;
            public Matrix viewMatrix;
            public Matrix projectionMatrix;
        }

        private struct DirectionalLight
        {
            public Vector2 rotation;
            public Vector3 direction;
            public Vector4 color;
            public Matrix viewMatrix;
            public Matrix projectionMatrix;
        }

        private struct Material
        {
            public Effect effect;
            public Vector4 ambient;
            public Vector4 diffuse;
        }

        private enum ShadowTechnique
        {
            None,
            ShadowMapping,
            ShadowMappingWith2x2PercentageCloserFiltering,
            ShadowMappingWith3x3PercentageCloserFiltering
        }

        private GraphicsDeviceManager graphics;
        private SpriteBatch spriteBatch;
        private SpriteFont spriteFont;
        private Vector2 fontPos;
        private KeyboardState currentKeyboardState;
        private KeyboardState prevKeyboardState;
        private MouseState currentMouseState;
        private MouseState prevMouseState;
        private Camera camera;
        private int windowWidth;
        private int windowHeight;
        private int frames;
        private int framesPerSecond;
        private TimeSpan elapsedTime = TimeSpan.Zero;
        private TimeSpan mouseIdleElapsedTime = TimeSpan.Zero;
        private bool displayHelp;
        private bool displayShadowMap;
        private bool displayShadows;
        private bool displaySceneFromLightSource;
        private ShadowTechnique shadowTechnique;
        private ShadowMap shadowMap;
        private DirectionalLight light;
        private Material material;
        private Model model;
        private BoundingSphere modelBounds;
        private Dictionary<int, Texture2D> modelTextures;
        
        public Demo()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            Window.Title = "XNA 4.0 Shadow Mapping Demo";
            IsMouseVisible = true;
            IsFixedTimeStep = false;

            displayHelp = false;
            displayShadowMap = false;
            displayShadows = true;
            displaySceneFromLightSource = false;

            shadowTechnique = ShadowTechnique.ShadowMappingWith3x3PercentageCloserFiltering;

            modelTextures = new Dictionary<int, Texture2D>();
        }

        private void CreateShadowMap()
        {
            shadowMap.Begin(GraphicsDevice);
            shadowMap.Draw(model, Matrix.Identity);
            shadowMap.End();
        }

        private string GetShadowMappingTechnique()
        {
            switch (shadowTechnique)
            {
            case ShadowTechnique.None:
                return "No shadows";

            case ShadowTechnique.ShadowMapping:
                return "Shadow mapping";

            case ShadowTechnique.ShadowMappingWith2x2PercentageCloserFiltering:
                return "Shadow mapping with 2x2 PCF";

            case ShadowTechnique.ShadowMappingWith3x3PercentageCloserFiltering:
                return "Shadow mapping with 3x3 PCF";

            default:
                return "No shadows";
            }
        }

        protected override void Initialize()
        {
            // Setup the window to be a quarter the size of the desktop.
            windowWidth = GraphicsDevice.DisplayMode.Width / 2;
            windowHeight = GraphicsDevice.DisplayMode.Height / 2;

            // Setup frame buffer.
            graphics.SynchronizeWithVerticalRetrace = false;
            graphics.PreferredBackBufferWidth = windowWidth;
            graphics.PreferredBackBufferHeight = windowHeight;
            graphics.PreferMultiSampling = true;
            graphics.ApplyChanges();

            // Position the text.
            fontPos = new Vector2(1.0f, 1.0f);

            // Setup the initial input states.
            currentKeyboardState = Keyboard.GetState();

            // Setup the shadow map.
            // The initial depth bias value looks good for this scene only.
            shadowMap = new ShadowMap(GraphicsDevice, Content);
            shadowMap.DepthBias = 0.013f;

            // Setup scene material.
            material.ambient = new Vector4(0.2f, 0.2f, 0.2f, 1.0f);
            material.diffuse = new Vector4(0.8f, 0.8f, 0.8f, 1.0f);

            // Setup scene light.
            light.direction = new Vector3(0.0f, -1.0f, -1.0f);
            light.direction.Normalize();
            light.color = new Vector4(1.0f, 1.0f, 1.0f, 1.0f);
            light.rotation = Vector2.Zero;

            base.Initialize();
        }

        private bool KeyJustPressed(Keys key)
        {
            return currentKeyboardState.IsKeyDown(key) && prevKeyboardState.IsKeyUp(key);
        }

        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);
            spriteFont = Content.Load<SpriteFont>(@"Fonts\DemoFont");
            
            // Setup effects.
            material.effect = Content.Load<Effect>(@"Effects\Lambert");

            // Setup model.
            LoadModel(@"Models\ShadowMappingScene", out model, out modelBounds);

            // Setup camera to look at the loaded model.
            camera.target = Vector3.Zero;
            camera.minZoom = modelBounds.Radius * 0.45f;
            camera.maxZoom = modelBounds.Radius;
            camera.offset = camera.maxZoom;
            camera.znear = 0.1f;
            camera.zfar = camera.maxZoom * 2.0f;
            camera.fovy = MathHelper.PiOver4;
            camera.orientation = Quaternion.CreateFromYawPitchRoll(0.0f, MathHelper.ToRadians(25.0f), 0.0f);
        }

        private void LoadModel(string assetName, out Model model, out BoundingSphere bounds)
        {
            model = Content.Load<Model>(assetName);
            bounds = new BoundingSphere();

            foreach (ModelMesh mesh in model.Meshes)
            {
                bounds = BoundingSphere.CreateMerged(bounds, mesh.BoundingSphere);

                foreach (ModelMeshPart part in mesh.MeshParts)
                {
                    BasicEffect basicEffect = part.Effect as BasicEffect;

                    if (basicEffect != null)
                        modelTextures[mesh.GetHashCode()] = basicEffect.Texture;

                    part.Effect = material.effect;
                }
            }
        }

        private void ProcessKeyboard()
        {
            prevKeyboardState = currentKeyboardState;
            currentKeyboardState = Keyboard.GetState();

            if (KeyJustPressed(Keys.Escape))
                this.Exit();

            if (KeyJustPressed(Keys.H))
                displayHelp = !displayHelp;

            if (currentKeyboardState.IsKeyDown(Keys.LeftAlt) ||
                currentKeyboardState.IsKeyDown(Keys.RightAlt))
            {
                if (KeyJustPressed(Keys.Enter))
                    ToggleFullScreen();
            }

            if (KeyJustPressed(Keys.Space))
                displayShadows = !displayShadows;

            if (KeyJustPressed(Keys.L))
                displaySceneFromLightSource = !displaySceneFromLightSource;
                        
            if (KeyJustPressed(Keys.S))
                displayShadowMap = !displayShadowMap;

            if (KeyJustPressed(Keys.Add))
                shadowMap.DepthBias = shadowMap.DepthBias + 0.001f;

            if (KeyJustPressed(Keys.Subtract))
            {
                shadowMap.DepthBias = shadowMap.DepthBias - 0.001f;

                if (shadowMap.DepthBias <= 0.0f)
                    shadowMap.DepthBias = 0.0f;
            }

            if (KeyJustPressed(Keys.D1))
                shadowTechnique = ShadowTechnique.ShadowMapping;

            if (KeyJustPressed(Keys.D2))
                shadowTechnique = ShadowTechnique.ShadowMappingWith2x2PercentageCloserFiltering;

            if (KeyJustPressed(Keys.D3))
                shadowTechnique = ShadowTechnique.ShadowMappingWith3x3PercentageCloserFiltering;
        }

        private void ProcessMouse(GameTime gameTime)
        {
            prevMouseState = currentMouseState;
            currentMouseState = Mouse.GetState();

            camera.translate.X = camera.translate.Y = 0.0f;
            camera.rotation.X = camera.rotation.Y = 0.0f;

            light.rotation.X = light.rotation.Y = 0.0f;

            float dx = 0.0f;
            float dy = 0.0f;
            float dz = 0.0f;

            // Process mouse button clicks.

            if (currentMouseState.LeftButton == ButtonState.Pressed)
            {
                dx = currentMouseState.X - prevMouseState.X;
                dx *= Camera.TRACKING_SPEED;

                dy = prevMouseState.Y - currentMouseState.Y;
                dy *= Camera.TRACKING_SPEED;

                camera.translate.X = dx;
                camera.translate.Y = dy;
            }
            else if (currentMouseState.MiddleButton == ButtonState.Pressed)
            {
                dz = currentMouseState.Y - prevMouseState.Y;
                dz *= Camera.DOLLYING_SPEED;

                camera.offset += dz;
            }
            else if (currentMouseState.RightButton == ButtonState.Pressed)
            {
                dx = currentMouseState.X - prevMouseState.X;
                dx *= Camera.ROTATION_SPEED;

                dy = currentMouseState.Y - prevMouseState.Y;
                dy *= Camera.ROTATION_SPEED;

                if (currentKeyboardState.IsKeyDown(Keys.LeftControl) ||
                    currentKeyboardState.IsKeyDown(Keys.RightControl))
                {
                    light.rotation.X = dy;
                    light.rotation.Y = dx;
                }
                else
                {
                    camera.rotation.X = dy;
                    camera.rotation.Y = dx;
                }
            }

            // Process mouse wheel scrolling.

            if (currentMouseState.ScrollWheelValue > prevMouseState.ScrollWheelValue)
                camera.offset -= Camera.MOUSEWHEEL_SPEED;
            else if (currentMouseState.ScrollWheelValue < prevMouseState.ScrollWheelValue)
                camera.offset += Camera.MOUSEWHEEL_SPEED;

            // Limit zooming.
                        
            if (camera.offset <= camera.minZoom)
                camera.offset = camera.minZoom;
            else if (camera.offset >= camera.maxZoom)
                camera.offset = camera.maxZoom;

            // Hide the mouse cursor when the mouse hasn't moved.

            int deltaX = currentMouseState.X - prevMouseState.X;
            int deltaY = currentMouseState.Y - prevMouseState.Y;
            bool idle = deltaX == 0 && deltaY == 0;

            if (idle)
            {
                mouseIdleElapsedTime += gameTime.ElapsedGameTime;

                if (mouseIdleElapsedTime > TimeSpan.FromSeconds(2.0))
                    IsMouseVisible = false;
            }
            else
            {
                IsMouseVisible = true;
                mouseIdleElapsedTime = TimeSpan.Zero;
            }
        }

        private void ToggleFullScreen()
        {
            int newWidth = 0;
            int newHeight = 0;

            graphics.IsFullScreen = !graphics.IsFullScreen;

            if (graphics.IsFullScreen)
            {
                newWidth = GraphicsDevice.DisplayMode.Width;
                newHeight = GraphicsDevice.DisplayMode.Height;
            }
            else
            {
                newWidth = windowWidth;
                newHeight = windowHeight;
            }

            graphics.PreferredBackBufferWidth = newWidth;
            graphics.PreferredBackBufferHeight = newHeight;
            graphics.ApplyChanges();
        }

        protected override void Update(GameTime gameTime)
        {
            if (!IsActive)
                return;

            ProcessKeyboard();

            if (!displaySceneFromLightSource)
                ProcessMouse(gameTime);

            UpdateCamera();
            UpdateLight();
            UpdateShadowMap();
            UpdateFrameRate(gameTime);
            
            base.Update(gameTime);
        }

        private void UpdateCamera()
        {
            // Calculate the view matrix.

            Quaternion rotation = Quaternion.Identity;
            float heading = MathHelper.ToRadians(camera.rotation.Y);
            float pitch = MathHelper.ToRadians(camera.rotation.X);

            if (heading != 0.0f)
            {
                rotation = Quaternion.CreateFromAxisAngle(Vector3.UnitY, heading);
                Quaternion.Concatenate(ref rotation, ref camera.orientation, out camera.orientation);
            }

            if (pitch != 0.0f)
            {
                rotation = Quaternion.CreateFromAxisAngle(Vector3.UnitX, pitch);
                Quaternion.Concatenate(ref camera.orientation, ref rotation, out camera.orientation);
            }

            Matrix.CreateFromQuaternion(ref camera.orientation, out camera.viewMatrix);

            Vector3 xAxis = new Vector3(camera.viewMatrix.M11, camera.viewMatrix.M21, camera.viewMatrix.M31);
            Vector3 yAxis = new Vector3(camera.viewMatrix.M12, camera.viewMatrix.M22, camera.viewMatrix.M32);
            Vector3 zAxis = new Vector3(camera.viewMatrix.M13, camera.viewMatrix.M23, camera.viewMatrix.M33);

            camera.target -= xAxis * camera.translate.X;
            camera.target -= yAxis * camera.translate.Y;

            camera.position = camera.target + zAxis * camera.offset;

            camera.viewMatrix.M41 = -Vector3.Dot(xAxis, camera.position);
            camera.viewMatrix.M42 = -Vector3.Dot(yAxis, camera.position);
            camera.viewMatrix.M43 = -Vector3.Dot(zAxis, camera.position);

            Vector3.Negate(ref zAxis, out camera.viewDir);

            // Calculate the projection matrix.

            camera.projectionMatrix = Matrix.CreatePerspectiveFieldOfView(camera.fovy,
                (float)windowWidth / (float)windowHeight, camera.znear, camera.zfar);
        }

        private void UpdateFrameRate(GameTime gameTime)
        {
            elapsedTime += gameTime.ElapsedGameTime;

            if (elapsedTime > TimeSpan.FromSeconds(1))
            {
                elapsedTime -= TimeSpan.FromSeconds(1);
                framesPerSecond = frames;
                frames = 0;
            }
        }

        private void IncrementFrameCounter()
        {
            ++frames;
        }

        private void UpdateLight()
        {
            Quaternion rotation = Quaternion.Identity;
            Quaternion temp = Quaternion.Identity;
            float heading = MathHelper.ToRadians(light.rotation.Y);
            float pitch = MathHelper.ToRadians(light.rotation.X);

            if (heading != 0.0f)
            {
                temp = Quaternion.CreateFromAxisAngle(Vector3.UnitY, heading);
                Quaternion.Concatenate(ref temp, ref rotation, out rotation);
            }

            if (pitch != 0.0f)
            {
                temp = Quaternion.CreateFromAxisAngle(Vector3.UnitX, pitch);
                Quaternion.Concatenate(ref rotation, ref temp, out rotation);
            }

            light.direction = Vector3.Transform(light.direction, rotation);
            light.direction.Normalize();

            if (displaySceneFromLightSource)
            {
                light.viewMatrix = Matrix.CreateLookAt(
                    light.direction * -camera.maxZoom,
                    Vector3.Zero,
                    Vector3.Up);

                light.projectionMatrix = Matrix.CreatePerspectiveFieldOfView(
                    camera.fovy,
                    (float)windowWidth / (float)windowHeight,
                    camera.znear,
                    camera.zfar * 3.0f);
            }
        }

        private void UpdateShadowMap()
        {
            Matrix viewProjection = camera.viewMatrix * camera.projectionMatrix;
            shadowMap.Update(light.direction, viewProjection);
        }
                
        private void DrawModel(Model model)
        {
            foreach (ModelMesh m in model.Meshes)
            {
                foreach (Effect e in m.Effects)
                {
                    if (!displayShadows)
                    {
                        e.CurrentTechnique = e.Techniques["Lambert"];
                    }
                    else
                    {
                        switch (shadowTechnique)
                        {
                        case ShadowTechnique.ShadowMapping:
                            e.CurrentTechnique = e.Techniques["LambertWithShadows"];
                            break;

                        case ShadowTechnique.ShadowMappingWith2x2PercentageCloserFiltering:
                            e.CurrentTechnique = e.Techniques["LambertWithShadows2x2PCF"];
                            e.Parameters["texelSize"].SetValue(shadowMap.TexelSize);
                            break;

                        case ShadowTechnique.ShadowMappingWith3x3PercentageCloserFiltering:
                            e.CurrentTechnique = e.Techniques["LambertWithShadows3x3PCF"];
                            e.Parameters["texelSize"].SetValue(shadowMap.TexelSize);
                            break;
                        }

                        e.Parameters["lightViewProjection"].SetValue(shadowMap.LightViewProjectionMatrix);
                        e.Parameters["textureScaleBias"].SetValue(shadowMap.TextureScaleBiasMatrix);
                        e.Parameters["depthBias"].SetValue(shadowMap.DepthBias);
                        e.Parameters["shadowMap"].SetValue(shadowMap.ShadowMapTexture);
                    }

                    e.Parameters["world"].SetValue(Matrix.Identity);

                    if (displaySceneFromLightSource)
                    {
                        e.Parameters["view"].SetValue(light.viewMatrix);
                        e.Parameters["projection"].SetValue(light.projectionMatrix);
                    }
                    else
                    {
                        e.Parameters["view"].SetValue(camera.viewMatrix);
                        e.Parameters["projection"].SetValue(camera.projectionMatrix);
                    }

                    e.Parameters["lightDir"].SetValue(light.direction);
                    e.Parameters["lightColor"].SetValue(light.color);
                    e.Parameters["materialAmbient"].SetValue(material.ambient);
                    e.Parameters["materialDiffuse"].SetValue(material.diffuse);
                    e.Parameters["colorMap"].SetValue(modelTextures[m.GetHashCode()]);
                }

                m.Draw();
            }
        }

        private void DrawShadowMap()
        {
            Rectangle rect = new Rectangle();

            rect.X = 0;
            rect.Y = GraphicsDevice.Viewport.Height - 128;
            rect.Width = 128;
            rect.Height = 128;

            spriteBatch.Begin();
            spriteBatch.Draw(shadowMap.ShadowMapTexture, rect, Color.White);
            spriteBatch.End();
        }

        private void DrawText()
        {
            StringBuilder buffer = new StringBuilder();

            if (displayHelp)
            {
                buffer.AppendLine("Left mouse button and drag to track the camera");
                buffer.AppendLine("Middle mouse button and drag to dolly the camera");
                buffer.AppendLine("Right mouse button and drag to orbit the camera");
                buffer.AppendLine("Mouse wheel to dolly the camera");
                buffer.AppendLine();
                buffer.AppendLine("CTRL and right mouse button and drag to orbit the light");
                buffer.AppendLine();
                buffer.AppendLine("Press 1 to enable shadows with no percentage-closer filtering");
                buffer.AppendLine("Press 2 to enable shadows with 2x2 percentage-closer filtering");
                buffer.AppendLine("Press 3 to enable shadows with 3x3 percentage-closer filtering");
                buffer.AppendLine();
                buffer.AppendLine("Press L to enable/disable viewing of scene from light");
                buffer.AppendLine("Press S to show/hide shadow map");
                buffer.AppendLine("Press SPACE to enable/disable shadows");
                buffer.AppendLine("Press ALT and ENTER to toggle full screen");
                buffer.AppendLine("Press ESCAPE to exit");
                buffer.AppendLine();
                buffer.AppendLine("Press H to hide help");
            }
            else
            {
                buffer.AppendFormat("FPS: {0}", framesPerSecond);
                buffer.AppendLine();
                buffer.AppendFormat("Depth bias: {0}", shadowMap.DepthBias.ToString("f3"));
                buffer.AppendLine();

                if (displayShadows)
                    buffer.AppendLine(GetShadowMappingTechnique());

                buffer.AppendLine();
                buffer.AppendLine("Press H to display help");
            }

            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.AlphaBlend);
            spriteBatch.DrawString(spriteFont, buffer.ToString(), fontPos, Color.Yellow);
            spriteBatch.End();
        }

        protected override void Draw(GameTime gameTime)
        {
            if (!IsActive)
                return;

            if (displayShadows)
                CreateShadowMap();

            GraphicsDevice.Clear(Color.CornflowerBlue);
            GraphicsDevice.BlendState = BlendState.Opaque;
            GraphicsDevice.DepthStencilState = DepthStencilState.Default;
            GraphicsDevice.SamplerStates[0] = SamplerState.LinearWrap;
            GraphicsDevice.SamplerStates[1] = SamplerState.LinearWrap;

            DrawModel(model);

            if (displayShadowMap)
                DrawShadowMap();

            DrawText();
            base.Draw(gameTime);
            IncrementFrameCounter();
        }
    }
}